% This is the solution for tan(gamma). As only function of
% dz/dm and dz/dh. Not alpha; therefore, it's the more general
% solution. There are two solutions. Jeff's paper or mine at
% SEP-119.
clear all

clear z_der_gen
clear z_sect_gen 
clear hx_sect_gen
clear mx_sect_gen
clear dz_dm_nomin_phase_sect dz_dh_nomin_phase_sect
clear dz_dm_sect_der dz_dh_sect_der

clear z_sect_der 
clear mx_sect_der
clear hx_sect_der

closereq

addpath /net/koko/homes/sep/daniel/Research/current/Matlab/pswei

cwv=1;

fprintf('Beginning: backscatter-ps =%d\n',cwv);

%g=[0.9 1.001 1.3]
g=2.0

sg=(g-1)./(g+1)

mx_cut=0.;
hx_cut=0;

flag=-1;
show_surf=0;
show_ang=0;

bbgreen=[0,1,0];
bbred=[1,0,0];
bbblue=[0,0,1];
bbcyan=[0,1,1];
bbmagenta=[1,0,1];
bbyellow=[1,1,0];

hx_data=.5;
%hx_data=.0
mx_data=0.;
t_data=2.;

n_gamma=159;
gamma_0=-79;
d_gamma=1;

n_alpha=159;
alpha_0=-79;
d_alpha=1;

for i_alpha=1:n_alpha
  for i_gamma=1:n_gamma
    alpha(i_gamma,i_alpha)=alpha_0+(i_alpha-1)*d_alpha;
    gamma(i_gamma,i_alpha)=gamma_0+(i_gamma-1)*d_gamma;
  end
end

fprintf('Doing: PS ratio=%d\n',g);

vp_mod=1;
vs_mod=vp_mod/g;

sp_par_gen(1)=1/(1*vp_mod);
ss_par_gen(1)=1/(1*vs_mod);

% Compute surface and its borders
[z_ang_gen,mx_ang_gen,hx_ang_gen]=gen_surf_ang_cwv(@s_cwv,sp_par_gen,ss_par_gen,t_data,mx_data,hx_data,gamma,alpha,flag);

[min_z_ang,max_z_ang,min_mx_ang,max_mx_ang,min_hx_ang,max_hx_ang]=max_min_zmx(z_ang_gen,mx_ang_gen,hx_ang_gen,flag);

% Compute tangent plane analytically
[dz_dh,dz_dm]=gen_tan_plane_ang_cwv(@s_cwv,sp_par_gen,ss_par_gen,t_data,mx_data,hx_data,gamma,alpha,cwv,flag);

fprintf('Cut analytical tangent plane with phase angles at hx=%f \n',hx_cut)
[mx_sect_der,hx_sect_der,dz_dm_sect_der]=cut_section(mx_ang_gen,hx_ang_gen,dz_dm,'hx',hx_cut);
fprintf('Cut analytical tangent plane dz_dh at hx=%f \n',hx_cut)
[mx_sect_der,hx_sect_der,dz_dh_sect_der]=cut_section(mx_ang_gen,hx_ang_gen,dz_dh,'mx',mx_cut);

% Compute derivatives numerically
disp('Compute numerical dz/dm')
  [mx_sect_gen,hx_sect_gen,z_sect_gen,dz_dm_der_num]=cut_section_der(mx_ang_gen,hx_ang_gen,z_ang_gen,'hx',hx_cut);
disp('Compute numerical dz/dh')
  [mx_sect_gen,hx_sect_gen,z_sect_gen,dz_dh_der_num]=cut_section_der(mx_ang_gen,hx_ang_gen,z_ang_gen,'mx',mx_cut);
  
h1=mesh(mx_ang_gen,hx_ang_gen,dz_dh);
view(60,10)
min_mx_ang=-1;
max_mx_ang=1;
min_hx_ang=-.5;
max_hx_ang=1.5;

[min_dzdh,max_dzdh,min_mx_ang,max_mx_ang,min_hx_ang,max_hx_ang]=max_min_zmx(dz_dh,mx_ang_gen,hx_ang_gen,flag);
axis_imp_ang=[min_mx_ang,max_mx_ang,min_hx_ang,max_hx_ang,min_dzdh,max_dzdh]

zlabel('dz/dh');
xlabel('Midpoint');
ylabel('Offset');

hold on

hplot=plot3(mx_sect_gen,hx_sect_gen,dz_dh_der_num);
set (hplot,'LineWidth',4);

print -depsc2 ./Fig/ang_cwv_wei_surf.ps

%for i=1:159
%  AA=sg*(dz_dm_der_num(i))-((sg*sg)*dz_dh_der_num(i));
%  if AA==0
%   tgt1(i) = 0;
%  else
%   BB=1-sg^2;
%   CC=sg*dz_dm_der_num(i)-dz_dh_der_num(i);
%   tgt1(i)=(-BB+sqrt(BB^2-4*AA*CC))/(2*AA);
%  end
%end
%tgt2(ig,:)=(-B-sqrt(B^2-4*A(ig,:).*C(ig,:)))./(2*A(ig,:));
%tgt=dz_dh;

%if(show_ang)
%  fprintf('Im showing angle')
%  [min_z_sect,max_z_sect,min_mx_sect,max_mx_sect,min_hx_sect,max_hx_sect]=max_min_zmx(atand(dz_dm_sect_der),mx_sect_der,hx_sect_der,flag);
%else
%  [min_z_sect,max_z_sect,min_mx_sect,max_mx_sect,min_hx_sect,max_hx_sect]=max_min_zmx(dz_dm_sect_der,mx_sect_der,hx_sect_der,flag);
%end

%axis_der_ang=[min_mx_ang,max_mx_ang,min_hx_ang,max_hx_ang,min_z_sect,1.3*max_z_sect];


%figure
%hplot=plot(hx_sect_der(1,:),atand(real(tgt1(1,:))))
%set (hplot,'LineWidth',4);
%set (hplot,'Color',bbred);
%hold on
%hplot=plot(hx_sect_der(2,:),atand(tgt1(2,:)))
%set (hplot,'LineWidth',2);
%set (hplot,'Color',bbgreen);
%hplot=plot(hx_sect_der(3,:),atand(tgt1(3,:)))
%set (hplot,'LineWidth',2);
%set (hplot,'Color',bbblue);
%legend('0.9','1.001','1.3')
%xlabel('Hx (km)');
%ylabel('g(o)');
%title('1st Square root for analytical Sol. for different PS rations')

%print -depsc2 ./Fig/ang_cwv_2sols_1.ps

%figure
%hplot=plot(hx_sect_der(1,:),atand(real(tgt2(1,:))))
%set (hplot,'LineWidth',4);
%set (hplot,'Color',bbred);
%hold on
%hplot=plot(hx_sect_der(2,:),atand(tgt2(2,:)))
%set (hplot,'LineWidth',2);
%set (hplot,'Color',bbgreen);
%hplot=plot(hx_sect_der(3,:),atand(tgt2(3,:)))
%set (hplot,'LineWidth',2);
%set (hplot,'Color',bbblue);
%legend('0.9','1.001','1.3')
%xlabel('Hx (km)');
%ylabel('g(o)');
%title('2nd Square root for analytical Sol. for different PS rations')
%
%print -depsc2 ./Fig/ang_cwv_2sols_2.ps

%tgt=dz_dh_sect_der;

