module adcig_
  !! offset to angle conversion

  use utilities
  use fftwflags
  use pctrist_
  use mymath_

  implicit none  
  integer, parameter, private:: dbl=selected_int_kind(15)

  !!----------------------------------------------------------------
  !! FFT
  real,              private :: scaleF, scale1I, scale2I
  integer(kind=dbl), private ::  planF,  plan1I,  plan2I

  complex, dimension(:,:,:,:), allocatable, private :: cf
  complex, dimension(:,:,:),       pointer, private :: sftF  !! shared
  complex, dimension(:),           pointer, private :: sft1I !! shared
  complex, dimension(:),           pointer, private :: sft2I !! shared

  !!----------------------------------------------------------------
  !! angle conversion
  real,    dimension(:),   allocatable, private :: khi,kai,kzi,kmi
  complex, dimension(:,:), allocatable, private :: model, data
  real,    dimension(:,:), allocatable, private :: str
  real,    dimension(:,:,:), allocatable, private :: stretch

  real,    private :: eps, gamma
  logical, private :: push
  !!----------------------------------------------------------------

contains

  !----------------------------------------------------------------
  subroutine adcig_fft_init(nth,ni)
    integer, intent(in) :: nth
    integer, dimension(3), intent(in) :: ni
    integer :: stat
    scaleF  = 1./sqrt(1.*product(ni))
    scale1I = 1./sqrt(1.*ni(1))
    scale2I = 1./sqrt(1.*ni(3))

    allocate(cf(ni(1),ni(2),ni(3),nth), stat=stat)
    if(stat /=0) call seperr("cant allocate cf")

    call fftwnd_f77_create_plan(planF,3,ni   ,FFTW_FORWARD  &
    ,    FFTW_ESTIMATE + FFTW_THREADSAFE + FFTW_IN_PLACE)

    call fftwnd_f77_create_plan(plan1I,1,ni(1),FFTW_BACKWARD &
    ,    FFTW_ESTIMATE + FFTW_THREADSAFE + FFTW_IN_PLACE)

    call fftwnd_f77_create_plan(plan2I,1,ni(3),FFTW_BACKWARD &
    ,    FFTW_ESTIMATE + FFTW_THREADSAFE + FFTW_IN_PLACE)

  end subroutine adcig_fft_init
  !----------------------------------------------------------------
  subroutine adcig_F_fft(ith,isl,f)
    integer,           intent(in) :: ith
    integer(kind=dbl), intent(in) :: isl
    complex, dimension(:,:,:,:), pointer :: f

    cf(:,:,:,ith) = f(:,:,:,isl)
    call fftwnd_f77_one(planF,cf(:,:,:,ith),0) !! Xform in place
    f(:,:,:,isl) = cf(:,:,:,ith) * scaleF

  end subroutine adcig_F_fft
  !----------------------------------------------------------------
  subroutine adcig_I_fft(ith,isl,f)
    integer,           intent(in) :: ith
    integer(kind=dbl), intent(in) :: isl
    complex, dimension(:,:,:,:), pointer :: f
    integer :: n1,n2,n3
    integer :: i1,i2,i3

    n1=size(f,1)
    n2=size(f,2)
    n3=size(f,3)

    cf(:,:,:,ith) = f(:,:,:,isl)
    do i2=1,n2
       do i3=1,n3
          call fftwnd_f77_one(plan1I,cf(:,i2,i3,ith),0) !! Xform in place
       end do
    end do
    cf(:,:,:,ith) = cf(:,:,:,ith) * scale1I

    do i1=1,n1
       do i2=1,n2
          call fftwnd_f77_one(plan2I,cf(i1,i2,:,ith),0) !! Xform in place
       end do
    end do
    f(:,:,:,isl) = cf(:,:,:,ith) * scale2I

  end subroutine adcig_I_fft
  !----------------------------------------------------------------
  subroutine adcig_fft_close()
    integer :: stat
    call fftwnd_f77_destroy_plan(planF)
    call fftwnd_f77_destroy_plan(plan1I)
    call fftwnd_f77_destroy_plan(plan2I)

    deallocate(cf, stat=stat)
    if(stat /=0) call seperr("cant deallocate cf")
  end subroutine adcig_fft_close
  !----------------------------------------------------------------
  subroutine adcig_shift_init(a1,a2,a3)
    type(paxis), intent(in):: a1,a2,a3

    integer             :: i1,i2,i3
    real                :: pi, argF, arg1I, arg2I
    integer             :: stat

    allocate(sftF(a1%n,a2%n,a3%n), stat=stat)
    if(stat /=0) call seperr("cant allocate sftF")
    allocate(sft1I(a1%n), stat=stat)
    if(stat /=0) call seperr("cant allocate sft1I")
    allocate(sft2I(a3%n), stat=stat)
    if(stat /=0) call seperr("cant allocate sft2I")

    sftF =cmplx(1.,0.)
    sft1I=cmplx(1.,0.)
    sft2I=cmplx(1.,0.)

    if(a1%o/=0. .or. a2%o/=0. .or. a3%o/=0 ) then

       pi = acos(-1.)
       do i1=1,a1%n
          do i2=1,a2%n
             do i3=1,a3%n

                argF  = 2*pi &
                * ( (i1-1.)/a1%n * a1%o/a1%d &
                +   (i2-1.)/a2%n * a2%o/a2%d &
                +   (i3-1.)/a3%n * a3%o/a3%d )

                arg1I = 2*pi &
                * ( (i1-1.)/a1%n * a1%o/a1%d )

                arg2I = 2*pi &
                * ( (i3-1.)/a3%n * a3%o/a3%d )

                sftF(i1,i2,i3) = cmplx(cos(argF),-(+1)*sin(argF))
                sft1I(i1     ) = cmplx(cos(arg1I),-(-1)*sin(arg1I))
                sft2I(i3     ) = cmplx(cos(arg2I),-(-1)*sin(arg2I))

             end do
          end do
       end do

    end if

  end subroutine adcig_shift_init
  !----------------------------------------------------------------
  subroutine adcig_F_shift(isl,f)
    integer(kind=dbl), intent(in) :: isl
    complex, dimension(:,:,:,:), pointer :: f
    f(:,:,:,isl) = f(:,:,:,isl) * sftF
  end subroutine adcig_F_shift
  !----------------------------------------------------------------
  subroutine adcig_I_shift(isl,f)
    integer(kind=dbl), intent(in) :: isl
    complex, dimension(:,:,:,:), pointer :: f
    integer :: n1,n2,n3
    integer :: i1,i2,i3

    n1=size(f,1)
    n2=size(f,2)
    n3=size(f,3)

    do i2=1,n2
       do i3=1,n3
        f(:,i2,i3,isl) = f(:,i2,i3,isl) * sft1I 
       end do
    end do

    do i1=1,n1
       do i2=1,n2
	f(i1,i2,:,isl) = f(i1,i2,:,isl) * sft2I
       end do
    end do

  end subroutine adcig_I_shift
  !----------------------------------------------------------------
  subroutine adcig_shift_close()
    integer :: stat
    deallocate(sft1I, stat=stat)
    if(stat /=0) call seperr("cant deallocate sft1I")
    deallocate(sft2I, stat=stat)
    if(stat /=0) call seperr("cant deallocate sft2I")
    deallocate(sftF, stat=stat)
    if(stat /=0) call seperr("cant deallocate sftF")
  end subroutine adcig_shift_close
  !----------------------------------------------------------------  

  !----------------------------------------------------------------
  subroutine adcig_ang_init(nth,az,ah,am,aa,eps_in,push_in,gamma_in)
    integer,     intent(in) :: nth
    type(paxis), intent(in) :: az,ah,aa,am
    real,        intent(in) :: eps_in, gamma_in
    logical,     intent(in) :: push_in
    type(paxis)             :: akh, akm
    integer :: iz,ih,ia,im
    real    :: kz,kh,ka,km
    real    :: A,B,C,TA,arg,gg
    real    :: pi
    integer :: stat

    push  = push_in
    eps   = eps_in
    gamma = gamma_in

    pi = acos(-1.)

    gg = (gamma-1)/(gamma+1)

    B = 1 - gg**2

    allocate( data(nth,ah%n), stat=stat)
    if(stat /=0) call seperr("cant allocate data")
    allocate(model(nth,aa%n), stat=stat)
    if(stat /=0) call seperr("cant allocate model")
    allocate(  str(nth,ah%n), stat=stat)
    if(stat /=0) call seperr("cant allocate str")

    allocate(khi(ah%n), stat=stat)
    if(stat /=0) call seperr("cant allocate khi")
    allocate(kai(aa%n), stat=stat)
    if(stat /=0) call seperr("cant allocate kai")
    allocate(kzi(az%n), stat=stat)
    if(stat /=0) call seperr("cant allocate kzi")
    allocate(kmi(am%n), stat=stat)
    if(stat /=0) call seperr("cant allocate kmi")

    !! kz
    do iz = 1,az%n/2
       kzi(iz       ) = (        iz-1)
       kzi(iz+az%n/2) = (-az%n/2+iz-1)
    end do
    kzi = 2*pi* kzi / (az%n*az%d)

    !! ka
    do ia = 1,aa%n
       kai(ia) =  aa%o + (ia-1) * aa%d
    end do

    !! kh
    do ih = 1,ah%n
       khi(ih) = (-ah%n/2+ih-1)
    end do
    khi = 2*pi* khi / (ah%n*ah%d)
    !! akh
    akh%o = khi(1)
    akh%d = 2*pi / (ah%n*ah%d)
    akh%n = ah%n

    !! km
    do im = 1,am%n
       kmi(im) = (-am%n/2+im-1)
    end do
    kmi = 2*pi* kmi / (am%n*am%d)
    !! akm
    akm%o = kmi(1)
    akm%d = 2*pi / (am%n*am%d)
    akm%n = am%n

    if(     push) then !! push
       allocate(stretch(az%n,ah%n,am%n), stat=stat)
       if(stat /=0) call seperr("cant allocate stretch")

       stretch=0.
       do iz=1,az%n
          kz = kzi(iz)
          if (kz/=0.) then
             do ih = 1,ah%n
                kh = khi(ih)
		do im = 1,am%n !! Ojo This ONLY work for 1 cmp
                    km = kmi(im)
		    A = (km/kz)*gg - (kh/kz)*(gg**2)
		    C = (km/kz)*gg - (kh/kz)
		    arg = (- B + real(sqrt((B**2) - 4*A*C)))/(2*A)
                    ka = - 180* atan(arg) /pi !* sign(kh/kz)
                    stretch(iz,ih,im) = ka
		end do
             end do
          else
             stretch(iz,:,:) = 90.
          end if
       end do
    else               !! pull
       write(0,*) "Doing pull"
       allocate(stretch(az%n,aa%n,am%n), stat=stat)
       if(stat /=0) call seperr("cant allocate stretch")

       stretch=0.
       do iz=1,az%n
          kz = kzi(iz)
	  if (kz/=0.) then
           do ia = 1,aa%n
             ka = pi* kai(ia) /180
	     do im = 1,am%n !! Ojo This ONLY work for 1 cmp
		km = kmi(im)
		TA = ((km/kz) - gg*tan(ka))/(1 + gg*(km/kz)*tan(ka))
                kh = - kz * (tan(ka) + gg*TA)/(1 - gg*tan(ka)*TA)
                stretch(iz,ia,im) = kh
	     end do
           end do
	  else
	   stretch(iz,:,:) = 90.
	  end if
       end do
    end if

    call pctrist_init(nth,akh,aa,push,eps)
  end subroutine adcig_ang_init
  !----------------------------------------------------------------
  subroutine adcig_ang_x(ith,isl,f,az,ahx,aax,amx)
    integer,           intent(in) :: ith
    integer(kind=dbl), intent(in) :: isl
    type(paxis), intent(in) :: az,ahx,aax,amx
    complex, dimension(:,:,:,:), pointer :: f

    integer :: ihx,iax,imx,iz
    real    :: khx,kax,    kz
    integer :: jhx

    jhx = ahx%n/2

    do iz=1,az%n
       do imx = 1,amx%n
          str(ith,:) = stretch(iz,:,imx)
          call pctrist_define(ith,str)
          data(ith,1    :jhx    )=f(iz,1+jhx:jhx+jhx,imx,isl)
          data(ith,1+jhx:jhx+jhx)=f(iz,1    :jhx    ,imx,isl)

          call pctrist_apply(ith,model,data)

          f(iz,:,imx,isl) = model(ith,:)

       end do
    end do

  end subroutine adcig_ang_x
  !----------------------------------------------------------------
  subroutine adcig_ang_y(ith,isl,f,az,ahy,aay,ahx)
    integer,           intent(in) :: ith
    integer(kind=dbl), intent(in) :: isl
    type(paxis), intent(in) :: az,ahy,aay,ahx
    complex, dimension(:,:,:,:), pointer :: f

    integer :: ihy,iay,ihx,iz
    real    :: khy,kay,    kz
    integer :: jhy

    jhy = ahy%n/2

    do iz=1,az%n
       str(ith,:) = stretch(iz,:,iz) !OJO THIS IS WRONG CHECK WITH ang_x
       call pctrist_define(ith,str)

       do ihx = 1,ahx%n
          data(ith,1    :jhy    )=f(iz,ihx,1+jhy:jhy+jhy,isl)
          data(ith,1+jhy:jhy+jhy)=f(iz,ihx,1    :jhy    ,isl)

          call pctrist_apply(ith,model,data)

          f(iz,ihx,:,isl) = model(ith,:)

       end do
    end do

  end subroutine adcig_ang_y
  !----------------------------------------------------------------  
  subroutine adcig_ang_close()
    integer :: stat
    call pctrist_close()

    deallocate(khi, stat=stat)
    if(stat /=0) call seperr("cant deallocate khi")
    deallocate(kzi, stat=stat)
    if(stat /=0) call seperr("cant deallocate kzi")
    deallocate(kai, stat=stat)
    if(stat /=0) call seperr("cant deallocate kai")
    deallocate(kmi, stat=stat)
    if(stat /=0) call seperr("cant deallocate kmi")

    deallocate(data, stat=stat)
    if(stat /=0) call seperr("cant deallocate data")
    deallocate(model, stat=stat)
    if(stat /=0) call seperr("cant deallocate model")
    deallocate(str, stat=stat)
    if(stat /=0) call seperr("cant deallocate str")
    deallocate(stretch, stat=stat)
    if(stat /=0) call seperr("cant deallocate stretch")

  end subroutine adcig_ang_close
  !----------------------------------------------------------------
end module adcig_
